// SuperPuperMandelbrot.cpp:      .
//

#include "stdafx.h"

cl_context Cont;
cl_command_queue Cmd;
cl_program Prog;
cl_kernel Kernel;
cl_mem BufData;
cl_mem BufNum;
cl_uint Size;

bool InitOCL()
{
	cl_uint NumPlatf;
	cl_int ErrCode = clGetPlatformIDs(0, 0, &NumPlatf);
	if (ErrCode != CL_SUCCESS)
	{
		printf("Can not get platform IDs!\n");
		return false;
	}

	cl_platform_id *PlatIDs = (cl_platform_id*)malloc(NumPlatf*sizeof(cl_platform_id));
	ErrCode = clGetPlatformIDs(NumPlatf, PlatIDs, 0);
	if (ErrCode != CL_SUCCESS)
	{
		printf("Can not get platform IDs!\n");
		return false;
	}

	printf("Available platforms:\n\n");
	for (cl_uint i = 0; i < NumPlatf; i++)
	{
		char Name[1024];
		ErrCode = clGetPlatformInfo(PlatIDs[i], CL_PLATFORM_NAME, 1024, Name, 0);
		if (ErrCode != CL_SUCCESS)
		{
			printf("%u - <Unknown>\n", i);
		}
		else
		{
			printf("%u - %s\n", i, Name);
		}
	}
	printf("\nSelect platform: ");
	cl_uint PlatfInd;
	scanf_s("%u", &PlatfInd);
	PlatfInd %= NumPlatf;

	cl_platform_id PID = PlatIDs[PlatfInd];
	free(PlatIDs);

	cl_uint NumDevs;
	ErrCode = clGetDeviceIDs(PID, CL_DEVICE_TYPE_ALL, 0, 0, &NumDevs);
	if (ErrCode != CL_SUCCESS)
	{
		printf("Can not get device IDs!\n");
		return false;
	}

	cl_device_id *DevIDs = (cl_device_id*)malloc(NumDevs*sizeof(cl_device_id));

	ErrCode = clGetDeviceIDs(PID, CL_DEVICE_TYPE_ALL, NumDevs, DevIDs, 0);
	if (ErrCode != CL_SUCCESS)
	{
		printf("Can not get device IDs!\n");
		return false;
	}

	printf("\nAvailable devices:\n\n");
	for (cl_uint i = 0; i < NumDevs; i++)
	{
		char Name[1024];

		ErrCode = clGetDeviceInfo(DevIDs[i], CL_DEVICE_NAME, 1024, Name, 0);
		if (ErrCode != CL_SUCCESS)
		{
			printf("%u - <Unknown>\n", i);
		}
		else
		{
			printf("%u - %s\n", i, Name);
		}
	}
	printf("\nSelect platform: ");
	cl_uint DevInd;
	scanf_s("%u", &DevInd);
	DevInd %= NumDevs;

	cl_device_id DID = DevIDs[DevInd];
	free(DevIDs);

	Cont = clCreateContext(0, 1, &DID, 0, 0, &ErrCode);
	if (ErrCode != CL_SUCCESS)
	{
		printf("Can not create context!\n");
		return false;
	}

	Cmd = clCreateCommandQueue(Cont, DID, 0, &ErrCode);
	if (ErrCode != CL_SUCCESS)
	{
		clReleaseContext(Cont);
		printf("Can not create cmd!\n");
		return false;
	}

	HANDLE hFile = CreateFile(L"Shader.cpp", GENERIC_READ, FILE_SHARE_READ, 0, OPEN_EXISTING, 0, 0);
	size_t fSize = GetFileSize(hFile, 0);
	HANDLE hMap = CreateFileMapping(hFile, 0, PAGE_READONLY, 0, 0, 0);
	char *ShaderText = (char*)MapViewOfFile(hMap, FILE_MAP_READ, 0, 0, 0);

	Prog = clCreateProgramWithSource(Cont, 1, (const char**)&ShaderText, &fSize, &ErrCode);
	if (ErrCode != CL_SUCCESS)
	{
		UnmapViewOfFile(ShaderText);
		CloseHandle(hMap);
		CloseHandle(hFile);
		clReleaseCommandQueue(Cmd);
		clReleaseContext(Cont);
		printf("Can not create program!\n");
		return false;
	}

	ErrCode = clBuildProgram(Prog, 1, &DID, 0, 0, 0);
	if (ErrCode != CL_SUCCESS)
	{
		UnmapViewOfFile(ShaderText);
		CloseHandle(hMap);
		CloseHandle(hFile);

		size_t LogLen;
		clGetProgramBuildInfo(Prog, DID, CL_PROGRAM_BUILD_LOG, 0, 0, &LogLen);

		char* Log = (char*)malloc(LogLen);
		clGetProgramBuildInfo(Prog, DID, CL_PROGRAM_BUILD_LOG, LogLen, Log, 0);

		printf("Can not build program!\nLog:\n%s\n\n", Log);

		free(Log);

		clReleaseCommandQueue(Cmd);
		clReleaseContext(Cont);
		return false;
	}

	UnmapViewOfFile(ShaderText);
	CloseHandle(hMap);
	CloseHandle(hFile);

	Kernel = clCreateKernel(Prog, "CalcMandel", &ErrCode);
	if (ErrCode != CL_SUCCESS)
	{
		clReleaseProgram(Prog);
		clReleaseCommandQueue(Cmd);
		clReleaseContext(Cont);
		printf("Can not create kernel!\n");
		return false;
	}

	printf("Enter picture size: ");
	scanf_s("%u", &Size);

	BufData = clCreateBuffer(Cont, CL_MEM_READ_WRITE, Size*Size, 0, &ErrCode);
	if (ErrCode != CL_SUCCESS)
	{
		clReleaseKernel(Kernel);
		clReleaseProgram(Prog);
		clReleaseCommandQueue(Cmd);
		clReleaseContext(Cont);
		printf("Can not create data buffer!\n");
		return false;
	}

	BufNum = clCreateBuffer(Cont, CL_MEM_WRITE_ONLY, sizeof(cl_uint), 0, &ErrCode);
	if (ErrCode != CL_SUCCESS)
	{
		clReleaseMemObject(BufData);
		clReleaseKernel(Kernel);
		clReleaseProgram(Prog);
		clReleaseCommandQueue(Cmd);
		clReleaseContext(Cont);
		printf("Can not create num buffer!\n");
		return false;
	}

	return true;
}

void DeInitOCL()
{
	clReleaseMemObject(BufNum);
	clReleaseMemObject(BufData);
	clReleaseKernel(Kernel);
	clReleaseProgram(Prog);
	clReleaseCommandQueue(Cmd);
	clReleaseContext(Cont);
}

void SaveBitmap(wchar_t fName[], unsigned char* Data)
{
	BITMAPFILEHEADER Bmpf;
	Bmpf.bfType = 0x4D42;
	Bmpf.bfSize = sizeof(BITMAPFILEHEADER)+sizeof(BITMAPINFOHEADER)+sizeof(RGBQUAD)*Size*Size;
	Bmpf.bfReserved1 = 0;
	Bmpf.bfReserved2 = 0;
	Bmpf.bfOffBits = sizeof(BITMAPFILEHEADER)+sizeof(BITMAPINFOHEADER);

	BITMAPINFOHEADER Bmpi;
	ZeroMemory(&Bmpi, sizeof(BITMAPINFOHEADER));

	Bmpi.biBitCount = 32;
	Bmpi.biCompression = BI_RGB;
	Bmpi.biHeight = Size;
	Bmpi.biPlanes = 1;
	Bmpi.biSize = sizeof(BITMAPINFOHEADER);
	Bmpi.biWidth = Size;

	RGBQUAD *OutData = (RGBQUAD*)malloc(Size*Size*sizeof(RGBQUAD));

	for (unsigned int i = 0; i < Size*Size; i++)
	{
		if (Data[i] == 0)
		{
			OutData[i].rgbBlue = 255;
			OutData[i].rgbGreen = 255;
			OutData[i].rgbRed = 255;
		}
		else
		{
			OutData[i].rgbBlue = 0;
			OutData[i].rgbGreen = 0;
			OutData[i].rgbRed = 0;
		}
		OutData[i].rgbReserved = 0;
	}

	HANDLE hFile = CreateFile(fName, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, 0, CREATE_ALWAYS, 0, 0);

	DWORD temp;
	WriteFile(hFile, &Bmpf, sizeof(BITMAPFILEHEADER), &temp, 0);
	WriteFile(hFile, &Bmpi, sizeof(BITMAPINFOHEADER), &temp, 0);
	WriteFile(hFile, OutData, Size*Size*sizeof(RGBQUAD), &temp, 0);

	CloseHandle(hFile);

	free(OutData);
}

int _tmain(int argc, _TCHAR* argv[])
{
	if (InitOCL())
	{
		unsigned char *Data = (unsigned char*)malloc(Size*Size);

		const float Border = 1.5f;
		const float Threshold = 3.0f;
		const cl_uint NumOfIter = 500;
		size_t GlobWorkSize[2] = { Size, Size };

		clSetKernelArg(Kernel, 0, sizeof(float), &Border);
		clSetKernelArg(Kernel, 1, sizeof(float), &Threshold);
		clSetKernelArg(Kernel, 2, sizeof(cl_uint), &NumOfIter);
		clSetKernelArg(Kernel, 3, sizeof(cl_uint), &Size);
		clSetKernelArg(Kernel, 4, sizeof(cl_mem), &BufData);
		//clSetKernelArg(Kernel, 5, sizeof(cl_mem), &BufNum);

		cl_int ErrCode = clEnqueueNDRangeKernel(Cmd, Kernel, 2, 0, GlobWorkSize, 0, 0, 0, 0);
		if (ErrCode != CL_SUCCESS)
		{
			printf("Can not execute program!\n");
		}

		ErrCode = clEnqueueReadBuffer(Cmd, BufData, CL_TRUE, 0, Size*Size, Data, 0, 0, 0);
		if (ErrCode != CL_SUCCESS)
		{
			printf("Can not read buffer!\n");
		}

		SaveBitmap(L"C:\\temp\\Mandel.bmp", Data);

		free(Data);
		DeInitOCL();
		printf("Success!\n");
	}
	printf("Enter random number...");
	int qwe;
	scanf_s("%d", &qwe);
	return 0;
}

